@@ -1,10 +1,28 @@
 #-*-perl-*-
 # This is a wrapper to the lchat.pl routines that make life easier
 # to do ftp type work.
-# Mostly by Lee McLoughlin <lmjm@doc.ic.ac.uk>
+#
+#
+#
+# Copyright (C) 1990 - 1998   Lee McLoughlin
+#
+# Permission to use, copy, and distribute this software and its
+# documentation for any purpose with or without fee is hereby granted,
+# provided that the above copyright notice appear in all copies and
+# that both that copyright notice and this permission notice appear
+# in supporting documentation.
+#
+# Permission to modify the software is granted, but not the right to
+# distribute the modified code.  Modifications are to be distributed
+# as patches to released version.
+#
+# This software is provided "as is" without express or implied warranty.
+#
+#
 # based on original version by Alan R. Martello <al@ee.pitt.edu>
 # And by A.Macpherson@bnr.co.uk for multi-homed hosts
 #
+#
 # Basic usage:
 #  require 'ftp.pl';
 #  $ftp_port = 21;
@@ -23,8 +41,11 @@
 #  &ftp'close();
 #
 #
-# $Id: ftp.pl,v 2.6 1994/06/06 18:37:37 lmjm Exp lmjm $
+# $Id: ftp.pl,v 2.9 1998/05/29 19:02:00 lmjm Exp lmjm $
 # $Log: ftp.pl,v $
+# Revision 2.9  1998/05/29 19:02:00  lmjm
+# Lots of changes.  See CHANGES since 2.8 file.
+#
 # Revision 2.6  1994/06/06  18:37:37  lmjm
 # Switched to lchat - a subset of chat.
 # Allow for 'remote help's need to parse the help strings in the continuations
@@ -69,7 +90,7 @@ $timeout_read = 3 * $timeout;
 # Timeout an open
 $timeout_open = $timeout;
 
-$version = '$Revision: 2.6 $';
+$version = '$Revision: 2.9 $';
 
 # This is a "global" it contains the last response from the remote ftp server
 # for use in error messages
@@ -104,7 +125,7 @@ $ftp'drop_on_421 = 1;
 # Name of a function to call on a pathname to map it into a remote
 # pathname.
 $mapunixout = '';
-$manunixin = '';
+$mapunixin = '';
 
 # This is just a tracing aid.
 $ftp_show = 0;
@@ -124,9 +145,23 @@ $ftp'use_pasv = 0;    # 0=no (default), 1=yes
 # Variable only used if proxying
 $proxy = $proxy_gateway = $proxy_ftp_port = '';
 
+# EXPERIMENTAL:
+# Used for skey password handling
+# (Normally set elsewhere - this is just a sensible default.)
+# Is expected to take count and code as arguments and prompt
+# for the secret key  with 'password:' on stdout and then print the password.
+$ftp'keygen_prog = '/usr/local/bin/key';
+
 # Uncomment to turn on lots of debugging.
 # &debug( 10 );
 
+# Limit how much data any one ftp'get can pull back
+# Negative values cause the size check to be skipped.
+$max_get_size = -1;
+
+# Where I am connected to.
+$connect_site = '';
+
 # &ftp'debug( debugging_level )
 # Turn on debugging ranging from 1 = some to 10 = everything
 sub ftp'debug
@@ -159,10 +194,10 @@ sub open_alarm
 sub timed_open
 {
 	local( $site, $ftp_port, $retry_call, $attempts ) = @_;
-	local( $connect_site, $connect_port );
+	local( $connect_port );
 	local( $ret );
 
-	alarm( $timeout_open );
+	&alarm( $timeout_open );
 
 	while( $attempts-- ){
 		if( $ftp_show ){
@@ -263,7 +298,7 @@ sub ftp'open
 	$SIG{ 'ALRM' } = "ftp\'open_alarm";
 
 	local( $ret ) = eval "&timed_open( '$site', $ftp_port, $retry_call, $attempts )";
-	alarm( 0 );
+	&alarm( 0 );
 	$SIG{ 'ALRM' } = $old_sig;
 
 	if( $@ =~ /^timeout/ ){
@@ -290,15 +325,18 @@ sub ftp'login
 
 	if( $proxy ){
 		# Should site or real_site be used here?
-		&send( "USER $remote_user@$real_site" );
+		&send( "USER $remote_user\@$real_site" );
 	}
 	else {
 		&send( "USER $remote_user" );
 	}
+
+	# Loop to ignore any remote banner (from proxy)
 	$ret = &expect( $timeout,
-		2, 1,   # $remote_user logged in
-		331, 2,   # send password for $remote_user
-		332, 2 ); # account for login - not yet supported
+		       2, 1,   # $remote_user logged in
+		       331, 2,   # send password for $remote_user
+		       332, 2 ); # account for login - not yet supported
+
 	if( $ret == 99 ){
 		&service_closed();
 		$ret = 0;
@@ -309,6 +347,24 @@ sub ftp'login
 	}
 	elsif( $ret == 2 ){
 		# A password is needed
+
+		# check for s/key challenge - eg, [s/key 994 ph29005]
+		# If we are talking to skey then use remote_password as the
+		# secret to generate a real password
+		if( $ftp'response =~ m#\[s/key (\d+) (\w+)\]# ){
+			local( $count, $code ) = ($1, $2);
+
+			# TODO: report open failure & remove need for echo
+			open( SKEY, "echo $remote_password | $ftp'keygen_prog $count $code |" );
+			while( <SKEY> ){
+				if( ! /password:/ ){
+					chop( $remote_password = $_ );
+				}
+			}
+			close SKEY;
+			print $showfd "skey pass: $remote_password\n";
+		}
+
 		&send( "PASS $remote_password" );
 
 		$ret = &expect( $timeout,
@@ -398,7 +454,7 @@ sub ftp'cwd
 # &pasv()
 # Gets: nothing
 # Returns: nothing
-# Assumptions: yuo are connecting to an ftp server that implements PASV.
+# Assumptions: you are connecting to an ftp server that implements PASV.
 # The PASV is necessary when using SOCKS and firewalls because the firewall
 # acts as a proxy.
 sub pasv
@@ -430,20 +486,25 @@ sub pasv
 		return 0;
 	}
 	if( $ret == 1 ) {
-		$response =~ m/^227 Entering Passive Mode \((\d+),(\d+),(\d+),(\d+),(\d+),(\d+)\)$/;
-		$newhost = sprintf( "%d.%d.%d.%d", $1, $2, $3, $4 );
-		$newport = $5 * 256 + $6;
+		if( $response =~ m/^227 Entering Passive Mode \((\d+),(\d+),(\d+),(\d+),(\d+),(\d+)\)/ ){
+			$newhost = sprintf( "%d.%d.%d.%d", $1, $2, $3, $4 );
+			$newport = $5 * 256 + $6;
+		}
+		else {
+			print $showfd "Cannot parse passive response\n" if $ftp_show;
+			return 0;
+		}
 	}
 
 	# now need to connect() the new socket
 	if( ! &chat'open_newport( $newhost, $newport, *S2 ) ){
 		if( $retry_call ){
-			print $showfd "Failed to connect\n" if $ftp_show;
+			print $showfd "Failed to connect newport\n" if $ftp_show;
 			next;
 		}
 		else {
 			print $showfd "proxy connection failed " if $proxy;
-			print $showfd "Cannot open ftp to $connect_site\n" if $ftp_show;
+			print $showfd "Cannot open pasv ftp to $connect_site\n" if $ftp_show;
 			return 0;
 		}
 	}
@@ -467,7 +528,7 @@ sub ftp'dir_open
 	if( ! &open_data_socket() ){
 		return 0;
 	}
-	
+
 	if( $use_pasv ){
 		&pasv();
 	}
@@ -496,7 +557,10 @@ sub ftp'dir_open
 		*NS = S2;
 	}
 	else {
-		accept( NS, S ) || die "accept failed $!";
+		if( &accept( 'NS', 'S' ) < 0 ){
+			&close_data_socket;
+			return 0;
+		}
 	}
 
 	# 
@@ -570,7 +634,7 @@ sub read_alarm
 # Support for read
 sub timed_read
 {
-	alarm( $timeout_read );
+	&alarm( $timeout_read );
 
 	return sysread( NS, $ftpbuf, $ftpbufsize );
 }
@@ -583,7 +647,7 @@ sub read_nosel
 	}
 
 	local( $ret ) = eval '&timed_read()';
-	alarm( 0 );
+	&alarm( 0 );
 
 	if( $@ =~ /^timeout/ ){
 		return -1;
@@ -607,6 +671,22 @@ sub read
 	return sysread( NS, $ftpbuf, $ftpbufsize );
 }
 
+sub write
+{
+	if( ! $service_open ){
+		return -1;
+	}
+
+	vec( $write_out, fileno( NS ), 1 ) = 1;
+
+	($nfound, $timeleft) = select( undef, $write_in = $write_out, undef, $to = $timeout_read );
+
+	if( $nfound <= 0 ){
+		return -1;
+	}
+	return syswrite( NS, $ftpbuf, $ftpbufsize );
+}
+
 # &ftp'dostrip( true or false )
 # Turn on or off stripping of incoming carriage returns.
 sub ftp'dostrip
@@ -682,13 +762,19 @@ sub ftp'get
 		*NS = S2;
 	}
 	else {
-		accept( NS, S ) || die "accept failed $!";
+		if( &accept( 'NS', 'S' ) < 0 ){
+			&close_data_socket;
+			return 0;
+		}
 	}
 
 	# 
 	# the data should be coming at us now
 	#
 
+	# for systems that differentiate between text and binary.
+	eval "binmode( NS )";
+
 	#
 	#  open the local fname
 	#  concatenate on the end if restarting, else just overwrite
@@ -700,6 +786,11 @@ sub ftp'get
 
 		return 0;
 	}
+	# Make sure the file can be updated - but only by me!
+	chmod( 0644, $loc_fname );
+
+	# for systems that differentiate between text and binary.
+	eval "binmode( FH )";
 
 	local( $start_time ) = time;
 	local( $bytes, $lasthash, $hashes ) = (0, 0, 0);
@@ -713,9 +804,15 @@ sub ftp'get
 #	while( ($len = &read_nosel()) > 0 ){
 
 # If you have select() then use the following line.
+	local( $too_big ) = 0;
 	while( ($len = &read()) > 0 ){
 
 		$bytes += $len;
+		if( $max_get_size > 0 && $bytes > $max_get_size ){
+			$too_big = 1;
+			$bytes = -1;
+			last;
+		}
 		if( $strip_cr ){
 			$ftpbuf =~ s/\r//g;
 		}
@@ -739,17 +836,21 @@ sub ftp'get
 # Add the next line back if you don't have select().
 #	$SIG{ 'ALRM' } = $old_sig;
 
-	close( FH );
-
 	# shut down our end of the socket
 	&close_data_socket;
 
+	if( ! close( FH ) ){
+		print $showfd "\nclose of local file failed: $!";
+
+		return 0;
+	}
+
 	if( $len < 0 ){
 		print $showfd "\ntimed out reading data!\n";
 
 		return 0;
 	}
-		
+
 	if( $ftp_show && $bytes > 0 ){
 		if( $hashes && ($hashes % $hashnl) != 0 ){
 			print $showfd "\n";
@@ -763,6 +864,11 @@ sub ftp'get
 		print $showfd "Got $bytes bytes ($rate bytes/sec)\n";
 	}
 
+	if( $too_big ){
+		print $showfd "Transfer exceeded allowed limit of $max_get_size\
+	";
+	}
+
 	#
 	# read the close
 	#
@@ -819,7 +925,7 @@ sub deldir
 
 # &ftp'put( local filename, remote filename, restart where left off )
 # Similar to get but sends file to the remote site.
-sub put
+sub ftp'put
 {
 	local( $loc_fname, $rem_fname ) = @_;
 	local( $strip_cr );
@@ -870,7 +976,10 @@ sub put
 		*NS = S2;
 	}
 	else {
-		accept( NS, S ) || die "accept failed $!";
+		if( &accept( 'NS', 'S' ) < 0 ){
+			&close_data_socket;
+			return 0;
+		}
 	}
 
 	# 
@@ -882,24 +991,90 @@ sub put
 	#
 	if( !open( FH, "<$loc_fname" ) ){
 		print $showfd "Cannot open local file $loc_fname\n";
-
+ 
 		# shut down our end of the socket
 		&close_data_socket;
 
 		return 0;
 	}
+
+	# for systems that differentiate between text and binary.
+	eval "binmode( FH )";
+
+	#while( <FH> ){
+	#	if( ! $service_open ){
+	#		last;
+	#	}
+	#	print NS ;
+	#}
+	#close( FH );
 	
-	while( <FH> ){
-		if( ! $service_open ){
-			last;
+	local( $bytes_written, $bytes, $lasthash, $hashes ) = (0, 0, 0, 0);
+
+# read the data from FH
+	while( ($len = sysread( FH, $ftpbuf, $ftpbufsize )) > 0 ){
+		#check size?
+		$bytes += $len;
+
+	 	if( $max_get_size > 0 && $bytes > $max_get_size ){
+                        $too_big = 1;
+                        $bytes = -1;
+                        last;
+                }
+		# write the data to NS checking that all data is written
+		# if syswrite returns 0 - written all of file
+		$left = $len ;
+		while( ($len2 = &write( NS, $ftpbuf, $left )) > 0 ){
+			$left = $len - $len2 ;
+			$bytes_written+= $len2;
+			$ftpbuf = substr( $ftpbuf, $len2, $left);
+		}
+		if( $len2 < 0 ){
+			print "error occurred while writing to network\n";
+			return 0;
+		}
+
+		# if ( $ftp_show ){ print out some hashes }
+                if( $ftp_show ){
+                        while( $bytes > ($lasthash + $hashevery) ){
+                                print $showfd '#';
+                                $lasthash += $hashevery;
+                                $hashes++;
+                                if( ($hashes % $hashnl) == 0 ){
+                                        print $showfd "\n";
+                                }
+                        }
 		}
-		print NS ;
 	}
-	close( FH );
+
+	if( $ret < 0 ){
+		print "error occurred while reading from $loc_fname\n";
+		return 0;
+	}
+	if( $bytes_written != $bytes ){
+		# This should never happen but ...
+		print "number of bytes written not equal to number read\n";
+		exit 0;
+	}
+ 	if( $ftp_show && $bytes > 0 ){
+                if( $hashes && ($hashes % $hashnl) != 0 ){
+                        print $showfd "\n";
+                }
+                local( $secs ) = (time - $start_time);
+                if( $secs <= 0 ){
+                        $secs = 1; # To avoid a divide by zero;
+                }
+
+                local( $rate ) = int( $bytes / $secs );
+                print $showfd "Got $bytes bytes ($rate bytes/sec)\n";
+        }
+
+        if( $too_big ){
+                print $showfd "Transfer exceeded allowed limit of $max_get_size\n";
+        }
 	
 	# shut down our end of the socket to signal EOF
 	&close_data_socket;
-	
 	#
 	# read the close
 	#
@@ -1031,6 +1206,10 @@ sub ftp'pwd
 		if( $response =~ /^2\d\d\s*"(.*)"\s.*$/ ){
 			$cwd = $1;
 		}
+		# For VMS
+		elsif( $response =~ /^2\d\d\ (.*) is the current directory/ ){
+			$cwd = $1;
+		}
 	}
 	return $cwd;
 }
@@ -1205,12 +1384,12 @@ sub expect
 	if( $drop_on_421 ){
 		# Handle 421 specially - has to go first in case a pattern
 		# matches on a generic 4.. response
-		push( @expect_args, "[.|\n]*^(421 .*)\\015\\n" );
+		push( @expect_args, "[.|\n]*^(421 .*)\\015?\\n" );
 		push( @expect_args, "&expectgot( \$1, 0 ); 99" );
 	}
 
 	# Match any obvious continuations.
-	push( @expect_args, "[.|\n]*^(\\d\\d\\d-.*|[^\\d].*)\\015\\n" );
+	push( @expect_args, "[.|\n]*^(\\d\\d\\d-.*|[^\\d].*)\\015?\\n" );
 	push( @expect_args, "&expectgot( \$1, 0 ); 100" );
 
 	while( @_ ){
@@ -1224,16 +1403,16 @@ sub expect
 			$pre = "[.|\n]*^";
 			$post = '\d\d ';
 		}
-		push( @expect_args, "$pre(" . $code . $post . ".*)\\015\\n" );
+		push( @expect_args, "$pre(" . $code . $post . ".*)\\015?\\n" );
 		push( @expect_args,
 			"&expectgot( \$1, 0 ); " . shift( @_ ) );
 	}
 	# Match any numeric response codes not explicitly looked for.
-	push( @expect_args, "[.|\n]*^(\\d\\d\\d .*)\\015\\n" );
+	push( @expect_args, "[.|\n]*^(\\d\\d\\d .*)\\015?\\n" );
 	push( @expect_args, "&expectgot( \$1, 0 ); 0" );
 	
 	# Treat all unrecognised lines as continuations
-	push( @expect_args, "^(.*)\\015\\n" );
+	push( @expect_args, "^(.*)\\015?\\n" );
 	push( @expect_args, "&expectgot( \$1, 0 ); 100" );
 	
 	# add patterns TIMEOUT and EOF
@@ -1260,6 +1439,10 @@ sub expect
 #
 sub open_data_socket
 {
+	if( $use_pasv ){
+		return 1;
+	}
+
 	local( $sockaddr, $port );
 	local( $type, $myaddr, $a, $b, $c, $d );
 	local( $mysockaddr, $family, $hi, $lo );
@@ -1269,8 +1452,14 @@ sub open_data_socket
 	($a,$b,$c,$d) = unpack( 'C4', $chat'thisaddr );
 	$this = $chat'thisproc;
 	
-	socket( S, $main'pf_inet, $main'sock_stream, $main'tcp_proto ) || die "socket: $!";
-	bind( S, $this ) || die "bind: $!";
+	if( ! socket( S, $main'pf_inet, $main'sock_stream, $main'tcp_proto ) ){
+		warn "socket: $!";
+		return 0;
+	}
+	if( ! bind( S, $this ) ){
+		warn "bind: $!";
+		return 0;
+	}
 	
 	# get the port number
 	$mysockaddr = getsockname( S );
@@ -1283,7 +1472,10 @@ sub open_data_socket
 	# we MUST do a listen before sending the port otherwise
 	# the PORT may fail
 	#
-	listen( S, 5 ) || die "listen";
+	if( ! listen( S, 5 ) ){
+		warn "listen: $!";
+		return 0;
+	}
 	
 	&send( "PORT $a,$b,$c,$d,$hi,$lo" );
 	
@@ -1316,6 +1508,22 @@ sub send
 	&chat'print( "$send_cmd\r\n" );
 }
 
+sub accept
+{
+	local( $NS, $S ) = @_;
+	local( $nfound, $accept_in, $accept_out, $timeleft, $to );
+
+	vec( $accept_in, fileno( $S ), 1 ) = 1;
+
+	($nfound, $timeleft) = select( $accept_out = $accept_in, undef, undef, $to = $timeout_read );
+
+	if( $nfound <= 0 ){
+		return -1;
+	}
+	
+	return accept( $NS, $S );
+}
+
 sub printargs
 {
 	while( @_ ){
@@ -1335,6 +1543,12 @@ sub filesize
 	
 }
 
+sub alarm
+{
+	local( $time_to_sig ) = @_;
+	eval "alarm( $time_to_sig )";
+}
+
 # Reply codes, see RFC959:
 # 1yz Positive Preliminary.  Expect another reply before proceeding
 # 2yz Positive Completion.
