@@ -1,10 +1,28 @@
 #-*-perl-*-
+# Copyright (C) 1990 - 1998   Lee McLoughlin
+#
+# Permission to use, copy, and distribute this software and its
+# documentation for any purpose with or without fee is hereby granted,
+# provided that the above copyright notice appear in all copies and
+# that both that copyright notice and this permission notice appear
+# in supporting documentation.
+#
+# Permission to modify the software is granted, but not the right to
+# distribute the modified code.  Modifications are to be distributed
+# as patches to released version.
+#
+# This software is provided "as is" without express or implied warranty.
+#
 # Parse "ls -lR" type listings
 # use lsparse'reset( dirname ) repeately
-# By Lee McLoughlin <lmjm@doc.ic.ac.uk>
 #
-# $Id: lsparse.pl,v 2.7 1994/06/10 18:28:24 lmjm Exp lmjm $
+# By Lee McLoughlin <lmjm@icparc.ic.ac.uk>
+#
+# $Id: lsparse.pl,v 2.9 1998/05/29 19:04:19 lmjm Exp lmjm $
 # $Log: lsparse.pl,v $
+# Revision 2.9  1998/05/29 19:04:19  lmjm
+# Lots of changes.  See CHANGES since 2.8 file.
+#
 # Revision 2.7  1994/06/10  18:28:24  lmjm
 # Another netware variant.
 # Another dosish system.
@@ -51,6 +69,12 @@ $lsparse'vms_keep_case = '';
 # A name to report when errors occur
 $lsparse'name = 'unknown';
 
+# Wether to report subdirs when finding them in a directory
+# or when their details appear.  (If you report early then mirro might
+# recreate locally remote restricted directories.)
+$lsparse'report_subdir = 0;	# Report when finding details.
+
+
 # Name of routine to call to parse incoming listing lines
 $ls_line = '';
 
@@ -75,9 +99,12 @@ sub lsparse'reset
 # This calls the filestore specific parser.
 sub lsparse'line
 {
+	local( $fh ) = @_;
+
 	# ls_line is setup in lsparse'reset to the name of the function
 	local( $path, $size, $time, $type, $mode ) =
-		eval "&$ls_line( @_ )";
+		eval "&$ls_line( \$fh )";
+
 
 	# Zap any leading ./  (Somehow they still creep thru.)
 	$path =~ s:^(\./)+::;
@@ -162,8 +189,7 @@ sub lsparse'line_unix
 			$non_crud = "";
 		}
 		
-
-		if( /^([\-FlrwxsStTdDa]{10}).*\D(\d+)\s*(\w\w\w\s+\d+\s*(\d+:\d+|\d\d\d\d))\s+(.*)\n/ ){
+		if( /^([\-FlrwxsStTdDam]{10}).*\D(\d+)\s*([A-Za-z]{3}\s+\d+\s*(\d+:\d+|\d\d\d\d))\s+(.*)\n/ ){
 			local( $kind, $size, $lsdate, $file ) = ($1, $2, $3, $5);
 			
 			if( $file eq '.' || $file eq '..' ){
@@ -186,7 +212,7 @@ sub lsparse'line_unix
 			elsif( $kind =~ /^d/i ){
 				# Don't create private dirs when not
 				# using recurse_hard.
-				if( ! $main'recurse_hard ){
+				if( $report_subdirs ){
 					next;
 				}
 
@@ -735,7 +761,7 @@ sub lsparse'line_dosish
 		# I'm about to look at this at lot
 		study;
 
-		if( m,(\d+)-(\d+)-(\d+)\s+(\d+):(\d+)(AM|PM)\s+(\d+|<DIR>)\s+(\S+), ){
+		if( m,(\d+)-(\d+)-(\d+)\s+(\d+):(\d+)(AM|PM)\s+(\d+|<DIR>)\s+(\S.*), ){
 			local( $mon, $day, $yr, $hrs, $min, $ampm, $dir_or_size, $file ) =
 				($1, $2, $3, $4, $5, $6, $7, $8);
 			if( $file eq '.' || $file eq '..' ){
@@ -743,6 +769,10 @@ sub lsparse'line_dosish
 			}
 
 			$hrs += 12 if $ampm =~ /PM/;
+			if( $hrs == 12 || $hrs == 24 ){
+				$hrs -= 12;
+			}
+
 			# TODO: fix hacky 19$yr
 			local( $lsdate ) = "$day-$mon-19$yr $hrs:$min";
 			local( $time ) = &main'lstime_to_time( $lsdate );
@@ -758,6 +788,26 @@ sub lsparse'line_dosish
 			$file =~ s,/+,/,g;
 			return( substr( $file, 1 ), $size, $time, $type, $mode );
 		}
+		# Match starts of directories.
+		elsif( /^([\.\/\\]*.*):$/ ){
+			$dir = $1;
+			# Switch from dos to unix slashes.
+			$dir =~ s,\\,/,g;
+			if( $dir eq '.' ){
+				next;
+			}
+			elsif( $dir !~ /^\// ){
+				$currdir = "$here/$dir";
+			}
+			else {
+				$currdir = "$dir";
+			}
+			$currdir =~ s,/+,/,g;
+			$match = $currdir;
+			$match =~ s/([\+\(\)\[\]\*\?])/\\$1/g;
+			return( substr( $currdir, 1 ), 0, 0, 'd', 0 );
+		}
+
 		else {
 			printf( "Unmatched line: %s", $_ );
 		}
@@ -906,7 +956,7 @@ sub lsparse'line_os2
 		# I'm about to look at this at lot
 		study;
 
-		if( m,(\d+)\s+((\S+)\s+)?((\S+)\s+)?(\d+)-(\d+)-(\d+)\s+(\d+):(\d+)\s+(\S+), ){
+		if( m,(\d+)\s+((\S+)\s+)?((\S+)\s+)?(\d+)-(\d+)-(\d+)\s+(\d+):(\d+)\s+(\S.*), ){
 			local( $size, $flags, $dir, $mon, $day, $yr, $hrs, $min, $file ) =
 				($1, $3, $5, $6, $7, $8, $9, $10, $11);
 			if( $file eq '.' || $file eq '..' ){
@@ -924,7 +974,6 @@ sub lsparse'line_os2
 			local( $time ) = &main'lstime_to_time( $lsdate );
 			local( $type ) = ($dir eq 'DIR' ? 'd' : 'f');
 			local( $mode ) = 0;
-			local( $size ) = 0;
 
 			$size = $dir_or_size if $dir_or_size =~ /^\d/;
 
@@ -1173,6 +1222,53 @@ sub lsparse'line_infomac
 }
 
 
+# --------------------- EPLF by Dan Bernstein
+# +i8388621.48638,m848117771,r,s1336,     qmsmac.html
+# +i8388621.88705,m850544954,/,   txt
+#
+sub lsparse'line_eplf
+{
+	local( $fh ) = @_;
+
+	if( $lsparse'readtime ){
+		alarm( $lsparse'readtime );
+	}
+
+	if( eof( $fh ) ){
+		alarm( 0 );
+		return( "", 0, 0, 0 );
+	}
+
+	while( <$fh> ){
+		s/\015//g;
+
+		# Store listing
+		print main'STORE $_;
+
+# +i8388621.48638,m848117771,r,s1336,     qmsmac.html
+# +i8388621.88705,m850544954,/,   txt
+		if( ! m:^\+i(\d+\.\d+),m(\d+),(/|[rw],s(\d+)),\s+(.*)$: ){
+			printf( "Unmatched line: %s\n", $_ );
+			next;
+		}
+		local( $dev_ino, $time, $dirrw, $size, $file ) = ($1, $2, $3, $4, $5);
+		local( $mode );
+		if( $dirrw =~ m:^/: ){
+			$type = 'd';
+			$size = 0;
+			$mode = 0755;
+		}
+		else {
+			$type = 'f';
+			$mode = ($dirrw =~ /r/ ? 0444 : 0666 );
+		}
+		return( $file, $size, $time, $type, $mode );
+	}
+	alarm( 0 );
+	return( '', 0, 0, 0, 0 );
+}
+
+
 # --------------------- CTAN files list
 #    22670 Mon Jul 20 12:36:34 1992 pub/tex/biblio/bibtex/contrib/aaai-named.bst
 #
