@@ -1,6 +1,7 @@
 #!/usr/local/bin/perl -s
 
 unshift( @INC, '.' );
+unshift( @INC, '..' );
 require 'lsparse.pl';
 
 $debug = 1;
@@ -24,6 +25,9 @@ if( $netware ){
 if( $dosftp ){
 	$lsparse'fstype = 'dosftp';
 }
+if( $dosish ){
+	$lsparse'fstype = 'dosish';
+}
 if( $macos ){
 	$lsparse'fstype = 'macos';
 }
@@ -33,16 +37,22 @@ if( $infomac ){
 if( $ctan ){
 	$lsparse'fstype = 'ctan';
 }
+if( $danber ){
+	$lsparse'fstype = 'danber';
+}
+warn "parse type is: $lsparse'fstype\n";
 
 &lsparse'reset( $dir );
 
+open( STORE, ">-" ) || die "cannot open stdout: $!";
+
 while( !eof( LS ) ){
 	($path, $size, $time, $type, $mode ) = &lsparse'line( "main'LS" );
 	last if( !$path );
 	if( $debug ){
-		local( $s ) = $T ? &t2str( $time ) : ' ';
-		printf "local: %s %s %s %s %0o\n",
-			$path, $size, $s, "[$time]", $type, $mode;
+		local( $s ) = ($T ? &t2str( $time ) : ' ');
+		printf "local: %s, %s, %s [%s], %s, %0o\n",
+			$path, $size, $s, $time, $type, $mode;
 	}
 }
 close( LS );
@@ -59,5 +69,5 @@ sub t2str
 	local($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = @t;
 
 	return sprintf( "%02d/%02d/%02d-%02d:%02d:%02d",
-		$year, $mon + 1, $mday, $hour, $him, $sec );
+		$year, $mon + 1, $mday, $hour, $min, $sec );
 }
